@@ -143,17 +143,9 @@ func (n *nodeContext) insertComprehension(
 		ec = &envComprehension{
 			comp:   c,
 			vertex: n.node,
-
-			err:  nil,   // shut up linter
-			envs: nil,   // shut up linter
-			done: false, // shut up linter
 		}
 	}
 
-	if ec.done && len(ec.envs) == 0 {
-		return
-	}
-
 	x := c.Value
 
 	if !n.ctx.isDevVersion() {
@@ -161,9 +153,14 @@ func (n *nodeContext) insertComprehension(
 		ci.closeInfo.span |= ComprehensionSpan
 	}
 
+	node := n.node.DerefDisjunct()
+
 	var decls []Decl
 	switch v := ToExpr(x).(type) {
 	case *StructLit:
+		ci = n.splitStruct(v, ci)
+
+		kind := TopKind
 		numFixed := 0
 		var fields []Decl
 		for _, d := range v.Decls {
@@ -171,20 +168,32 @@ func (n *nodeContext) insertComprehension(
 			case *Field:
 				numFixed++
 
+				if f.Label.IsInt() {
+					kind &= ListKind
+				} else if f.Label.IsString() {
+					kind &= StructKind
+				}
+
 				// Create partial comprehension
 				c := &Comprehension{
 					Syntax:  c.Syntax,
 					Clauses: c.Clauses,
 					Value:   f,
-					arcType: f.ArcType,
+					arcType: f.ArcType, // TODO: can be derived, remove this field.
 
 					comp:   ec,
 					parent: c,
-					arc:    n.node,
+					arc:    node,
 				}
 
 				conjunct := MakeConjunct(env, c, ci)
-				n.node.state.insertFieldUnchecked(f.Label, ArcPending, conjunct)
+				if n.ctx.isDevVersion() {
+					n.assertInitialized()
+					n.insertArc(f.Label, ArcPending, conjunct, conjunct.CloseInfo, false)
+				} else {
+					n.insertFieldUnchecked(f.Label, ArcPending, conjunct)
+				}
+
 				fields = append(fields, f)
 
 			case *LetField:
@@ -200,12 +209,17 @@ func (n *nodeContext) insertComprehension(
 
 					comp:   ec,
 					parent: c,
-					arc:    n.node,
+					arc:    node,
 				}
 
 				conjunct := MakeConjunct(env, c, ci)
-				arc := n.node.state.insertFieldUnchecked(f.Label, ArcMember, conjunct)
-				arc.MultiLet = f.IsMulti
+				n.assertInitialized()
+				arc := n.insertFieldUnchecked(f.Label, ArcMember, conjunct)
+				if n.ctx.isDevVersion() {
+					arc.MultiLet = true
+				} else {
+					arc.MultiLet = f.IsMulti
+				}
 
 				fields = append(fields, f)
 
@@ -224,31 +238,45 @@ func (n *nodeContext) insertComprehension(
 			st := v
 			if len(fields) < len(v.Decls) {
 				st = &StructLit{
-					Src:   v.Src,
-					Decls: fields,
+					Src:             v.Src,
+					Decls:           fields,
+					isComprehension: true,
 				}
 			}
-			n.node.AddStruct(st, env, ci)
+			node.AddStruct(st, env, ci)
 			switch {
 			case !ec.done:
 				ec.structs = append(ec.structs, st)
 			case len(ec.envs) > 0:
-				st.Init()
+				st.Init(n.ctx)
+				if kind == StructKind || kind == ListKind {
+					n.updateNodeType(kind, st, ci)
+				}
 			}
 		}
 
+		c.kind = kind
+
 		switch numFixed {
 		case 0:
 			// Add comprehension as is.
 
 		case len(v.Decls):
 			// No comprehension to add at this level.
+			// The should be considered a struct if it has only non-regular
+			// fields (like definitions), and no embeddings.
+			if kind == TopKind {
+				c.kind = StructKind
+			}
 			return
 
 		default:
 			// Create a new StructLit with only the fields that need to be
 			// added at this level.
-			x = &StructLit{Decls: decls}
+			x = &StructLit{
+				Decls:           decls,
+				isComprehension: true,
+			}
 		}
 	}
 
@@ -288,7 +316,7 @@ func (c *OpContext) yield(
 		ctx:   c,
 		comp:  comp,
 		f:     f,
-		state: state.vertexStatus(),
+		state: state.status,
 	}
 	y := comp.Clauses[0]
 
@@ -402,7 +430,11 @@ func (n *nodeContext) processComprehension(d *envYield, state vertexStatus) *Bot
 			envs = append(envs, env)
 		}
 
-		if err := ctx.yield(d.vertex, d.env, d.comp, oldOnly(state), f); err != nil {
+		if err := ctx.yield(d.vertex, d.env, d.comp, combinedFlags{
+			status:    state,
+			condition: allKnown,
+			mode:      ignore,
+		}, f); err != nil {
 			if err.IsIncomplete() {
 				return err
 			}
@@ -421,7 +453,7 @@ func (n *nodeContext) processComprehension(d *envYield, state vertexStatus) *Bot
 
 		if len(d.envs) > 0 {
 			for _, s := range d.structs {
-				s.Init()
+				s.Init(n.ctx)
 			}
 		}
 		d.structs = nil
@@ -431,11 +463,13 @@ func (n *nodeContext) processComprehension(d *envYield, state vertexStatus) *Bot
 	d.inserted = true
 
 	if len(d.envs) == 0 {
+		n.node.updateArcType(ArcNotPresent)
 		return nil
 	}
 
 	v := n.node
 	for c := d.leaf; c.parent != nil; c = c.parent {
+		v = n.ctx.deref(v)
 		v.updateArcType(c.arcType)
 		if v.ArcType == ArcNotPresent {
 			parent := v.Parent
@@ -445,10 +479,19 @@ func (n *nodeContext) processComprehension(d *envYield, state vertexStatus) *Bot
 			ctx.current().state = taskFAILED
 			return nil
 		}
+		if k := c.kind; k == StructKind || k == ListKind {
+			v := v.DerefDisjunct()
+			if s := v.getBareState(n.ctx); s != nil {
+				s.updateNodeType(k, ToExpr(c.Value), d.id)
+			}
+		}
 		v = c.arc
 	}
 
 	id := d.id
+	// TODO: should we treat comprehension values as optional?
+	// It seems so, but it causes some hangs.
+	// id.setOptional(nil)
 
 	for _, env := range d.envs {
 		if n.node.ArcType == ArcNotPresent {
