@@ -684,7 +684,7 @@ func (n *nodeContext) detectCycleV3(arc *Vertex, env *Environment, x Resolver, c
 				//		c: a: int
 				//
 				// This is equivalent to a reference cycle.
-				if r.Depth == n.node.state.depth {
+				if r.Depth == n.depth {
 					return ci, true
 				}
 				ci.Refs = nil
@@ -1191,6 +1191,9 @@ func (n *nodeContext) reportCycleError() {
 		// TODO: probably, this should have the referenced arc.
 	}
 	n.setBaseValue(CombineErrors(nil, n.node.Value(), b))
+	// TODO(mem): might still be processing, so only use this when we
+	// exclude processed nodes.
+	// n.node.clearArcs(n.ctx)
 	n.node.Arcs = nil
 }
 
